/*
 * Copyright (C) 2011 Sébastien Villemot <sebastien.villemot@ens.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef _MODEL_SPEC_HH
#define _MODEL_SPEC_HH

#include <ostream>
#include <vector>

#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>

//! Generic class for describing a model specification
class ModelSpec
{
public:
  //! Number of endogenous
  const int ny;
  //! Number of exogenous
  const int nx;
  //! Number of state variables
  const int ns;
  //! Number of variables in forward part
  const int nf;

  ModelSpec(int ny_arg, int nx_arg, int ns_arg, int nf_arg);
  virtual ~ModelSpec();
  virtual void steady_state(gsl_vector *y_ss) const = 0;
  //! Computes all relative errors
  virtual void errors(const gsl_vector *y_prev, const gsl_vector *y_curr, const gsl_vector *fwd, const gsl_vector *shocks, gsl_vector *errors) const = 0;
  //! Computes the maximum error (in absolute value)
  double max_error(const gsl_vector *y_prev, const gsl_vector *y_curr, const gsl_vector *fwd, const gsl_vector *shocks) const;
  //! Computes the terms inside the expectancy operator of Euler equation
  /*! fwd is therefore an array of size nf */
  virtual void forward_part(const gsl_vector *y_prev, const gsl_vector *y_curr, const gsl_vector *y_next, const gsl_vector *shocks, gsl_vector *fwd) const = 0;
  //! Computes the steady state of the state vector (of length ns)
  virtual void state_steady_state(gsl_vector *s_ss) const = 0;
  //! Extract the previous period endogenous and shocks from state (for test 1)
  virtual void infer_from_state(const gsl_vector *state, gsl_vector *y_prev, gsl_vector *shocks) const = 0;
  //! Extract the state vector path from a simulation (for DHM stat)
  virtual void extract_state(const gsl_vector *init, const gsl_matrix *sims, gsl_matrix *state) const = 0;
  //! Write a one-line description of the specification
  virtual void write_output(std::ostream &output) const = 0;
  //! Write contents of the endogenous vector y in human-readable form
  virtual void write_state(std::ostream &output, const gsl_vector *y) const = 0;
  //! Write contents of the Euler errors vector errs in human-readable form
  virtual void write_errors(std::ostream &output, const gsl_vector *errs) const = 0;
  //! Names of state variables
  virtual void states_names(std::vector<std::string> &r) const = 0;
  //! Names of shocks (exogenous) variables
  virtual void shocks_names(std::vector<std::string> &r) const = 0;
  //! Names of all endogenous variables
  virtual void variables_names(std::vector<std::string> &r) const = 0;
  //! Names of Euler errors of all equations
  virtual void residuals_names(std::vector<std::string> &r) const = 0;
  //! Short name of the specification
  virtual std::string name() const = 0;
};

#endif
