/*
 * Copyright (C) 2009-2011 Sébastien Villemot <sebastien.villemot@ens.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <iostream>
#include <cassert>
#include <cmath>
#include <cstring>

#include <gsl/gsl_blas.h>
#include <gsl/gsl_deriv.h>

#include <matio.h>

#include "MultiCountryCGASolution.hh"

MultiCountryCGASolution::MultiCountryCGASolution(const MultiCountryRBCA2 &modspec_arg) throw (UnsupportedSpec)
  : ModelSolution(modspec_arg)
{
  const int n = modspec_arg.n;

  if (n != 2 && n != 4 && n != 6 && n != 8)
    throw UnsupportedSpec();

  // Retrieve MATLAB arrays
  char filename[50];
  snprintf(filename, 50, "cga/M6N%d.mat", n);

  mat_t *mfp = Mat_Open(filename, MAT_ACC_RDONLY);
  if (mfp == NULL)
    {
      std::cerr << "MultiCountryCGASolution::MultiCountryCGASolution: can't open " << filename << std::endl;
      exit(EXIT_FAILURE);
    }

  matvar_t *V_mx = Mat_VarReadInfo(mfp, (char *) "V");
  matvar_t *z0_mx = Mat_VarReadInfo(mfp, (char *) "z0");
  matvar_t *VI_mx = Mat_VarReadInfo(mfp, (char *) "VI");
  matvar_t *meanx_mx = Mat_VarReadInfo(mfp, (char *) "meanx");
  matvar_t *sigz0_mx = Mat_VarReadInfo(mfp, (char *) "sigz0");
  matvar_t *stdx_mx = Mat_VarReadInfo(mfp, (char *) "stdx");
  matvar_t *stdz_mx = Mat_VarReadInfo(mfp, (char *) "stdz");
  matvar_t *v_mx = Mat_VarReadInfo(mfp, (char *) "v");
  matvar_t *NC_mx = Mat_VarReadInfo(mfp, (char *) "NC");

  matvar_t *VIL_mx;

  VIL_mx = Mat_VarReadInfo(mfp, (char *) "VIL");

  // First deal with NC (needed for matrices dimensions)
  assert(NC_mx != NULL && NC_mx->class_type == MAT_C_DOUBLE && NC_mx->rank == 2
         && NC_mx->dims[0] == 1 && NC_mx->dims[1] == 1);
  Mat_VarReadDataAll(mfp, NC_mx);
  NC = (int) ((double *) NC_mx->data)[0];

  // Test array types and dimensions, and read data
  assert(V_mx != NULL && z0_mx != NULL && VI_mx != NULL && meanx_mx != NULL
         && sigz0_mx != NULL && stdx_mx != NULL && stdz_mx != NULL && v_mx != NULL);

  Mat_VarReadDataAll(mfp, V_mx);
  Mat_VarReadDataAll(mfp, z0_mx);
  Mat_VarReadDataAll(mfp, VI_mx);
  Mat_VarReadDataAll(mfp, meanx_mx);
  Mat_VarReadDataAll(mfp, sigz0_mx);
  Mat_VarReadDataAll(mfp, stdx_mx);
  Mat_VarReadDataAll(mfp, stdz_mx);
  Mat_VarReadDataAll(mfp, v_mx);

  assert(V_mx->class_type == MAT_C_DOUBLE && V_mx->rank == 2
         && V_mx->dims[0] == 2*n && V_mx->dims[1] == 2*n);
  assert(z0_mx->class_type == MAT_C_DOUBLE && z0_mx->rank == 2
         && z0_mx->dims[0] == NC && z0_mx->dims[1] == 2*n);
  assert(VI_mx->class_type == MAT_C_DOUBLE && VI_mx->rank == 3
         && (VI_mx->dims[0] == 2*n+1 || VI_mx->dims[0] == 1+2*n+2*n*(2*n+1)/2)
         && VI_mx->dims[1] == n && VI_mx->dims[2] == NC);
  assert(meanx_mx->class_type == MAT_C_DOUBLE && meanx_mx->rank == 2
         && meanx_mx->dims[0] == 1 && meanx_mx->dims[1] == 2*n);
  assert(sigz0_mx->class_type == MAT_C_DOUBLE && sigz0_mx->rank == 2
         && sigz0_mx->dims[0] == NC && sigz0_mx->dims[1] == 1);
  assert(stdx_mx->class_type == MAT_C_DOUBLE && stdx_mx->rank == 2
         && stdx_mx->dims[0] == 1 && stdx_mx->dims[1] == 2*n);
  assert(stdz_mx->class_type == MAT_C_DOUBLE && stdz_mx->rank == 2
         && stdz_mx->dims[0] == 1 && stdz_mx->dims[1] == 2*n);
  assert(v_mx->class_type == MAT_C_DOUBLE && v_mx->rank == 2
         && v_mx->dims[0] == 1 && v_mx->dims[1] == 1);

  assert(VIL_mx != NULL);
  Mat_VarReadDataAll(mfp, VIL_mx);
  VIL_dim2 = 1;
  assert(VIL_mx->class_type == MAT_C_DOUBLE && VIL_mx->rank == 3
         && (VIL_mx->dims[0] == 2*n+1 || VIL_mx->dims[0] == 1+2*n+2*n*(2*n+1)/2)
         && VIL_mx->dims[1] == VIL_dim2 && VIL_mx->dims[2] == NC);

  // Construct GSL structures
  V_prime = gsl_matrix_alloc(2*n, 2*n);
  gsl_matrix_memcpy(V_prime, &gsl_matrix_view_array((double *) V_mx->data, 2*n, 2*n).matrix);
  z0_prime = gsl_matrix_alloc(2*n, 3);
  gsl_matrix_memcpy(z0_prime, &gsl_matrix_view_array((double *) z0_mx->data, 2*n, 3).matrix);
  VI_dim1 = VI_mx->dims[0];
  VI = new double[VI_dim1*n*NC];
  memcpy(VI, VI_mx->data, VI_dim1*n*NC*sizeof(double));

  meanx = gsl_vector_alloc(2*n);
  gsl_vector_memcpy(meanx, &gsl_vector_view_array((double *) meanx_mx->data, 2*n).vector);
  sigz0 = gsl_vector_alloc(3);
  gsl_vector_memcpy(sigz0, &gsl_vector_view_array((double *) sigz0_mx->data, 3).vector);
  stdx = gsl_vector_alloc(2*n);
  gsl_vector_memcpy(stdx, &gsl_vector_view_array((double *) stdx_mx->data, 2*n).vector);
  stdz = gsl_vector_alloc(2*n);
  gsl_vector_memcpy(stdz, &gsl_vector_view_array((double *) stdz_mx->data, 2*n).vector);
  v = ((double *) v_mx)[0];

  VIL = new double[VI_dim1*VIL_dim2*NC];
  memcpy(VIL, VIL_mx->data, VI_dim1*VIL_dim2*NC*sizeof(double));

  // Free MX arrays
  Mat_VarFree(V_mx);
  Mat_VarFree(z0_mx);
  Mat_VarFree(VI_mx);
  Mat_VarFree(meanx_mx);
  Mat_VarFree(sigz0_mx);
  Mat_VarFree(stdx_mx);
  Mat_VarFree(stdz_mx);
  Mat_VarFree(v_mx);
  Mat_VarFree(NC_mx);
  Mat_VarFree(VIL_mx);

  Mat_Close(mfp);

  // Temporary variables
  xnorm = gsl_vector_alloc(2*n);
  znorm = gsl_vector_alloc(2*n);
  crbf = gsl_vector_alloc(NC);
  dist_z_z0 = gsl_vector_alloc(NC);
  rbfn = gsl_vector_alloc(NC);
  z1 = gsl_matrix_alloc(NC, 2*n);
  vi = gsl_matrix_alloc(VI_dim1, n);
  VI_tmp = gsl_matrix_alloc(VI_dim1, NC);
  instrs = gsl_vector_alloc(VI_dim1);
  ord2_vec = gsl_vector_alloc(2*n);
  ord2_mat = gsl_matrix_alloc(2*n, 2*n);

  vil = gsl_matrix_alloc(VI_dim1, VIL_dim2);
  VIL_tmp = gsl_matrix_alloc(VI_dim1, NC);
}

MultiCountryCGASolution::~MultiCountryCGASolution()
{
  gsl_matrix_free(V_prime);
  gsl_matrix_free(z0_prime);
  delete[] VI;
  gsl_vector_free(meanx);
  gsl_vector_free(sigz0);
  gsl_vector_free(stdx);
  gsl_vector_free(stdz);
  delete[] VIL;

  // Temporary variables
  gsl_vector_free(xnorm);
  gsl_vector_free(znorm);
  gsl_vector_free(crbf);
  gsl_vector_free(dist_z_z0);
  gsl_vector_free(rbfn);
  gsl_vector_free(ord2_vec);
  gsl_matrix_free(z1);
  gsl_matrix_free(vi);
  gsl_matrix_free(VI_tmp);
  gsl_matrix_free(ord2_mat);
  gsl_vector_free(instrs);

  gsl_matrix_free(vil);
  gsl_matrix_free(VIL_tmp);
}

void
MultiCountryCGASolution::policy_func_internal(const gsl_vector *y_prev, const gsl_vector *shocks, gsl_vector *y_curr)
{
  const MultiCountryRBCA2 &ms = dynamic_cast<const MultiCountryRBCA2 &>(modspec);

  const int n = ms.n;

  assert(y_prev->stride == 1 && y_curr->stride == 1);

  double *c_curr = gsl_vector_ptr(y_curr, 0);
  double *l_curr = gsl_vector_ptr(y_curr, n);
  double *i_curr = gsl_vector_ptr(y_curr, 2*n);
  double *k_curr = gsl_vector_ptr(y_curr, 3*n);
  double *a_curr = gsl_vector_ptr(y_curr, 4*n);

  const double *k_prev = gsl_vector_const_ptr(y_prev, 3*n);
  const double *a_prev = gsl_vector_const_ptr(y_prev, 4*n);

  // Set tech shock in y_curr
  for(int j = 0; j < n; j++)
    a_curr[j] = exp(ms.rho*log(a_prev[j]) + ms.sigma*(gsl_vector_get(shocks, j) + gsl_vector_get(shocks, n)));

  // Compute capital stocks
  for(int j = 0; j < n; j++)
    {
      gsl_vector_set(xnorm, j, k_prev[j]);
      gsl_vector_set(xnorm, j+n, a_curr[j]);
    }
  gsl_vector_sub(xnorm, meanx);
  gsl_vector_div(xnorm, stdx);

  gsl_blas_dgemv(CblasNoTrans, 1.0, V_prime, xnorm, 0.0, znorm);
  gsl_vector_div(znorm, stdz);

  gsl_vector_set_all(crbf, -0.5*v*v);
  gsl_vector_div(crbf, sigz0);
  gsl_vector_div(crbf, sigz0);

  for(int j = 0; j < NC; j++)
    gsl_matrix_set_row(z1, j, znorm);

  gsl_vector_set_zero(dist_z_z0);
  for(int j = 0; j < NC; j++)
    for(int k = 0; k < 2*n; k++)
      *gsl_vector_ptr(dist_z_z0, j) += gsl_matrix_get(z1, j, k) * gsl_matrix_get(z1, j, k) + gsl_matrix_get(z0_prime, k, j) * gsl_matrix_get(z0_prime, k, j) - 2 * gsl_matrix_get(z1, j, k) * gsl_matrix_get(z0_prime, k, j);

  for(int j = 0; j < NC; j++)
    gsl_vector_set(rbfn, j, exp(gsl_vector_get(crbf, j) * gsl_vector_get(dist_z_z0, j)));

  double sum = 0;
  for(int j = 0; j < NC; j++)
    sum += gsl_vector_get(rbfn, j);
  gsl_vector_scale(rbfn, 1/sum);

  for(int j = 0; j < n; j++)
    {
      for(int i = 0; i < VI_dim1; i++)
        for(int k = 0; k < NC; k++)
          gsl_matrix_set(VI_tmp, i, k, VI[i + VI_dim1*(j + n * k)]);
      gsl_blas_dgemv(CblasNoTrans, 1.0, VI_tmp, rbfn, 0.0, &gsl_matrix_column(vi, j).vector);
    }

  gsl_vector_set(instrs, 0, 1.0);
  gsl_vector_memcpy(&gsl_vector_subvector(instrs, 1, n).vector, &gsl_vector_const_view_array(k_prev, n).vector);
  gsl_vector_memcpy(&gsl_vector_subvector(instrs, n+1, n).vector, &gsl_vector_view_array(a_curr, n).vector);

  if (VI_dim1 > 2*n+1)
    {
      gsl_vector_memcpy(&gsl_vector_subvector(ord2_vec, 0, n).vector, &gsl_vector_const_view_array(k_prev, n).vector);
      gsl_vector_memcpy(&gsl_vector_subvector(ord2_vec, n, n).vector, &gsl_vector_view_array(a_curr, n).vector);
      gsl_matrix_set_zero(ord2_mat);
      gsl_blas_dsyr(CblasLower, 1.0, ord2_vec, ord2_mat);
      int k = 2*n+1;
      for(int j = 0; j < 2*n; j++)
        for(int i = j; i < 2*n; i++)
          gsl_vector_set(instrs, k++, gsl_matrix_get(ord2_mat, i, j));
    }
    
  gsl_blas_dgemv(CblasTrans, 1.0, vi, instrs, 0.0, &gsl_vector_view_array(k_curr, n).vector);

  // Set investment
  for(int j = 0; j < n; j++)
    i_curr[j] = k_curr[j] - (1.0-ms.delta)*k_prev[j];

  // Compute labor and consumption
  // Initialize labor
  for(int j = 0; j < VIL_dim2; j++)
    {
      for(int i = 0; i < VI_dim1; i++)
        for(int k = 0; k < NC; k++)
          gsl_matrix_set(VIL_tmp, i, k, VIL[i + VI_dim1*(j + VIL_dim2 * k)]);
      gsl_blas_dgemv(CblasNoTrans, 1.0, VIL_tmp, rbfn, 0.0, &gsl_matrix_column(vil, j).vector);
    }

  gsl_blas_ddot(instrs, &gsl_matrix_column(vil, 0).vector, l_curr); // Only labor supply of first country

  double dif = 1;

  // Loop to compute consumption and labor
  while(dif > 1e-10)
    {
      // Compute labor supply of other countries
      for(int j = 1; j < n; j++)
        l_curr[j] = pow(a_curr[j] * pow(k_prev[j], ms.alpha) / a_curr[0]
                        / pow(k_prev[0], ms.alpha) * ms.taus[0] * ms.b[0]
                        / ms.taus[j] / ms.b[j], ms.etas[j] / (1+ms.alpha*ms.etas[j]))
          * pow(l_curr[0], (ms.etas[j]+ms.alpha*ms.etas[j]*ms.etas[0])
                / (ms.etas[0]+ms.alpha*ms.etas[j]*ms.etas[0]));
      
      // Compute consumption of all countries
      for(int j = 0; j < n; j++)
        c_curr[j] = pow(ms.b[j]*pow(l_curr[j], 1/ms.etas[j])/ms.A/(1.0-ms.alpha)
                        / a_curr[j] / pow(k_prev[j], ms.alpha) / pow(l_curr[j], -ms.alpha),
                        -ms.gammas[j]);

      double el = 0;
      for(int j = 0; j < n; j++)
        el += c_curr[j] + k_curr[j] - k_prev[j] + ms.phi/2.0*pow(k_curr[j]/k_prev[j]-1.0, 2.0)*k_prev[j];
      for(int j = 1; j < n; j++)
        el -= ms.A * pow(k_prev[j], ms.alpha) * pow(l_curr[j], 1.0-ms.alpha) * a_curr[j];
      el /= ms.A * pow(k_prev[0], ms.alpha) * a_curr[0];
      el = pow(el, 1.0/(1.0-ms.alpha));

      l_curr[0] = 0.99*l_curr[0] + 0.01*el;
      dif = fabs(1.0-el/l_curr[0]);
    }

  // Set lambda to the mean of marginal utilites of consumption
  double uc_curr[n], ul_curr[n];
  gsl_vector_set(y_curr, 5*n, 0.0);
  ms.marginal_utilities(y_curr->data, uc_curr, ul_curr);
  for(int j = 0; j < n; j++)
    *gsl_vector_ptr(y_curr, 5*n) += uc_curr[j] * ms.taus[j];
  *gsl_vector_ptr(y_curr, 5*n) /= (double) n;
}

std::string
MultiCountryCGASolution::name() const
{
  return "CGA";
}
