/*
 * Copyright (C) 2010-2011 Sébastien Villemot <sebastien.villemot@ens.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <iostream>
#include <cassert>
#include <cmath>
#include <sstream>

#include <gsl/gsl_blas.h>
#include <gsl/gsl_linalg.h>

#include "MultiCountryHybSolution.hh"

MultiCountryHybSolution::MultiCountryHybSolution(const MultiCountryRBCA2 &modspec_arg, int order_arg, hybrid_t hybrid_arg) throw (UnsupportedSpec)
  : PerSolution(modspec_arg, matname(modspec_arg, order_arg), order_arg),
    n(modspec_arg.n), hybrid(hybrid_arg)
{
}

std::string
MultiCountryHybSolution::matname(const MultiCountryRBCA2 &modspec, int order)
{
  std::ostringstream s;
  s << "hyb/MC-RBC-A2-N" << modspec.n << "-O" << order << ".mat";
  return s.str();
}

MultiCountryHybSolution::~MultiCountryHybSolution()
{
}

void
MultiCountryHybSolution::create_state_vector(const gsl_vector *y_prev, const gsl_vector *shocks, gsl_vector *state) const
{
  // Create state vector (purely backward endo [k] + both backward and forward endo [a] + exo [e])
  // We take care of the fact that ys is in an different order than y_curr
  for(int j = 0; j < n; j++)
    {
      gsl_vector_set(state, j, gsl_vector_get(y_prev, 3*n+j) - gsl_vector_get(ys, n+j));
      gsl_vector_set(state, n+j, gsl_vector_get(y_prev, 4*n+j) - gsl_vector_get(ys, 2*n+j));
      gsl_vector_set(state, 2*n+j, gsl_vector_get(shocks, j));
    }
  gsl_vector_set(state, 3*n, gsl_vector_get(shocks, n));  
}

void
MultiCountryHybSolution::reorder_from_dynare(const gsl_vector *y_dynare, gsl_vector *y_curr) const
{
  // Move from internal order to declaration order (swap k with l, and a with i)
  for (size_t j = 0; j < (size_t) n; j++)
    {
      gsl_vector_set(y_curr, j, gsl_vector_get(y_dynare, j));
      gsl_vector_set(y_curr, n+j, gsl_vector_get(y_dynare, 3*n+j));
      gsl_vector_set(y_curr, 2*n+j, gsl_vector_get(y_dynare, 4*n+j));
      gsl_vector_set(y_curr, 3*n+j, gsl_vector_get(y_dynare, n+j));
      gsl_vector_set(y_curr, 4*n+j, gsl_vector_get(y_dynare, 2*n+j));
    }
  gsl_vector_set(y_curr, 5*n, gsl_vector_get(y_dynare, 5*n));
}

void
MultiCountryHybSolution::policy_func_internal(const gsl_vector *y_prev, const gsl_vector *shocks, gsl_vector *y_curr)
{
  policy_func_perturbation(y_prev, shocks, y_curr);

  const MultiCountryRBCA2 &ms = dynamic_cast<const MultiCountryRBCA2 &>(modspec);

  if (hybrid == Hybrid)
    {
      assert(y_prev->stride == 1 && y_curr->stride == 1);

      double *c_curr = gsl_vector_ptr(y_curr, 0);
      double *l_curr = gsl_vector_ptr(y_curr, n);
      double *k_curr = gsl_vector_ptr(y_curr, 3*n);
      double *a_curr = gsl_vector_ptr(y_curr, 4*n);

      const double *k_prev = gsl_vector_const_ptr(y_prev, 3*n);
      const double *a_prev = gsl_vector_const_ptr(y_prev, 4*n);

      // Compute tech shock exactly
      for (size_t j = 0; j < (size_t) n; j++)
        a_curr[j] = exp(ms.rho*log(a_prev[j]) + ms.sigma*(gsl_vector_get(shocks, j) + gsl_vector_get(shocks, n)));

      // Loop to compute consumption and labor
      double dif = 1;
      while (dif > 1e-10)
        {
          // Compute labor supply of other countries
          for (size_t j = 1; j < (size_t) n; j++)
            l_curr[j] = pow(a_curr[j] * pow(k_prev[j], ms.alpha) / a_curr[0]
                            / pow(k_prev[0], ms.alpha) * ms.taus[0] * ms.b[0]
                            / ms.taus[j] / ms.b[j], ms.etas[j] / (1+ms.alpha*ms.etas[j]))
              * pow(l_curr[0], (ms.etas[j]+ms.alpha*ms.etas[j]*ms.etas[0])
                    / (ms.etas[0]+ms.alpha*ms.etas[j]*ms.etas[0]));

          // Compute consumption of all countries
          for (size_t j = 0; j < (size_t) n; j++)
            c_curr[j] = pow(ms.b[j]*pow(l_curr[j], 1/ms.etas[j])/ms.A/(1.0-ms.alpha)
                            / a_curr[j] / pow(k_prev[j], ms.alpha) / pow(l_curr[j], -ms.alpha),
                            -ms.gammas[j]);

          double el = 0;
          for (size_t j = 0; j < (size_t) n; j++)
            el += c_curr[j] + k_curr[j] - k_prev[j] + ms.phi/2.0*pow(k_curr[j]/k_prev[j]-1.0, 2.0)*k_prev[j];
          for (size_t j = 1; j < (size_t) n; j++)
            el -= ms.A * pow(k_prev[j], ms.alpha) * pow(l_curr[j], 1.0-ms.alpha) * a_curr[j];
          el /= ms.A * pow(k_prev[0], ms.alpha) * a_curr[0];
          el = pow(el, 1.0/(1.0-ms.alpha));

          l_curr[0] = 0.99*l_curr[0] + 0.01*el;
          dif = fabs(1.0-el/l_curr[0]);
        }

      // Set lambda to the first marginal utility of consumption
      double uc_curr[n], ul_curr[n];
      gsl_vector_set(y_curr, 5*n, 0.0);
      ms.marginal_utilities(y_curr->data, uc_curr, ul_curr);
      *gsl_vector_ptr(y_curr, 5*n) = uc_curr[0] * ms.taus[0];
    }
}

std::string
MultiCountryHybSolution::name() const
{
  switch(order)
    {
    case 1:
      switch (hybrid)
        {
        case PlainPerturbation:
          return "PER1";
        case Hybrid:
          return "HYB1";
        }
    case 2:
      switch (hybrid)
        {
        case PlainPerturbation:
          return "PER2";
        case Hybrid:
          return "HYB2";
        }
    case 3:
      switch (hybrid)
        {
        case PlainPerturbation:
          return "PER3";
        case Hybrid:
          return "HYB3";
        }
    default:
      return "UNKNOWN";
    }
}
