/*
 * Copyright (C) 2011 Sébastien Villemot <sebastien.villemot@ens.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <R.h>
#include <Rinternals.h>

#include <sys/time.h>
#include <sys/resource.h>

#include <cstring>

#include <gsl/gsl_matrix.h>

#include "SolutionTester.hh"

#include "MultiCountryCGASolution.hh"
#include "MultiCountryHybSolution.hh"

double
timeval_substract(const struct timeval &x, const struct timeval &y)
{
  return ((double) (x.tv_sec - y.tv_sec))
    + ((double) (x.tv_usec - y.tv_usec)) / 1000000.0;
}

double
cputime(const struct rusage &t1, const struct rusage &t2)
{
  return round(100.0*(timeval_substract(t2.ru_utime, t1.ru_utime)
                    + timeval_substract(t2.ru_stime, t1.ru_stime)))/100.0;
}

extern "C"
{
  
  SEXP
  accuracy_test1(SEXP solution, SEXP nb_points, SEXP radius)
  {
    if (TYPEOF(solution) != EXTPTRSXP)
      error_return("First argument should be an external pointer");
    ModelSolution *sol = static_cast<ModelSolution *>(R_ExternalPtrAddr(solution));
    PROTECT(nb_points = coerceVector(nb_points, INTSXP));
    PROTECT(radius = coerceVector(radius, REALSXP));

    SolutionTester st(*sol, SolutionTester::GaussHermiteAndMonomialDegree5);

    const ModelSpec &spec = sol->modspec;

    // R stores matrices in column-major order, GSL in row-major order
    SEXP data, max_err, test_time;
    PROTECT(data = allocMatrix(REALSXP, spec.ns+2*spec.ny, *INTEGER(nb_points)));
    PROTECT(max_err = allocVector(REALSXP, 1));
    PROTECT(test_time = allocVector(REALSXP, 1));
  
    struct rusage r1, r2;
    getrusage(RUSAGE_SELF, &r1);
    st.accuracy_test1(*INTEGER(nb_points), *REAL(radius), *REAL(max_err),
                      &gsl_matrix_view_array(REAL(data), *INTEGER(nb_points), spec.ns+2*spec.ny).matrix);
    getrusage(RUSAGE_SELF, &r2);
    *REAL(test_time) = cputime(r1, r2);
    
    std::vector<std::string> rownames;
    spec.states_names(rownames);
    spec.variables_names(rownames);
    spec.residuals_names(rownames);
    SEXP rownames2;
    PROTECT(rownames2 = allocVector(STRSXP, rownames.size()));
    for (size_t i = 0; i < rownames.size(); i++)
      SET_STRING_ELT(rownames2, i, mkChar(rownames[i].c_str()));

    SEXP dimnames;
    PROTECT(dimnames = allocVector(VECSXP, 2));
    SET_VECTOR_ELT(dimnames, 0, rownames2);
    SET_VECTOR_ELT(dimnames, 1, R_NilValue);
    dimnamesgets(data, dimnames);

    SEXP r;
    PROTECT(r = allocVector(VECSXP, 3));
    SET_VECTOR_ELT(r, 0, max_err);
    SET_VECTOR_ELT(r, 1, data);
    SET_VECTOR_ELT(r, 2, test_time);

    SEXP names;
    PROTECT(names = allocVector(STRSXP, 3));
    SET_STRING_ELT(names, 0, mkChar("max.err"));
    SET_STRING_ELT(names, 1, mkChar("data"));
    SET_STRING_ELT(names, 2, mkChar("test.time"));
    namesgets(r, names);

    UNPROTECT(9);

    return r;
  }

  SEXP
  accuracy_test2(SEXP solution, SEXP T, SEXP warmup, SEXP rng_seed)
  {
    if (TYPEOF(solution) != EXTPTRSXP)
      error_return("First argument should be an external pointer");
    ModelSolution *sol = static_cast<ModelSolution *>(R_ExternalPtrAddr(solution));
    PROTECT(T = coerceVector(T, INTSXP));
    PROTECT(warmup = coerceVector(warmup, INTSXP));
    PROTECT(rng_seed = coerceVector(rng_seed, INTSXP));

    SolutionTester st(*sol, SolutionTester::GaussHermiteAndMonomialDegree5);

    const ModelSpec &spec = sol->modspec;

    // R stores matrices in column-major order, GSL in row-major order
    SEXP data, max_err, mean_err, test_time;
    PROTECT(data = allocMatrix(REALSXP, spec.nx+2*spec.ny, *INTEGER(T)));
    PROTECT(max_err = allocVector(REALSXP, 1));
    PROTECT(mean_err = allocVector(REALSXP, 1));
    PROTECT(test_time = allocVector(REALSXP, 1));

    struct rusage r1, r2;
    getrusage(RUSAGE_SELF, &r1);
    st.accuracy_test2(*INTEGER(T), *INTEGER(warmup), *REAL(max_err),
                      *REAL(mean_err), *INTEGER(rng_seed),
                      &gsl_matrix_view_array(REAL(data), *INTEGER(T), spec.nx+2*spec.ny).matrix);
    getrusage(RUSAGE_SELF, &r2);
    *REAL(test_time) = cputime(r1, r2);

    std::vector<std::string> rownames;
    spec.variables_names(rownames);
    spec.shocks_names(rownames);
    spec.residuals_names(rownames);
    SEXP rownames2;
    PROTECT(rownames2 = allocVector(STRSXP, rownames.size()));
    for (size_t i = 0; i < rownames.size(); i++)
      SET_STRING_ELT(rownames2, i, mkChar(rownames[i].c_str()));

    SEXP dimnames;
    PROTECT(dimnames = allocVector(VECSXP, 2));
    SET_VECTOR_ELT(dimnames, 0, rownames2);
    SET_VECTOR_ELT(dimnames, 1, R_NilValue);
    dimnamesgets(data, dimnames);

    SEXP r;
    PROTECT(r = allocVector(VECSXP, 4));
    SET_VECTOR_ELT(r, 0, max_err);
    SET_VECTOR_ELT(r, 1, mean_err);
    SET_VECTOR_ELT(r, 2, data);
    SET_VECTOR_ELT(r, 3, test_time);

    SEXP names;
    PROTECT(names = allocVector(STRSXP, 4));
    SET_STRING_ELT(names, 0, mkChar("max.err"));
    SET_STRING_ELT(names, 1, mkChar("mean.err"));
    SET_STRING_ELT(names, 2, mkChar("data"));
    SET_STRING_ELT(names, 3, mkChar("test.time"));
    namesgets(r, names);
    
    UNPROTECT(11);

    return r;
  }

  void
  solution_finalizer(SEXP ptr)
  {
    if (TYPEOF(ptr) != EXTPTRSXP)
      {
        Rf_error("Argument should be an external pointer");
        return;
      }
    ModelSolution *sol = static_cast<ModelSolution *>(R_ExternalPtrAddr(ptr));
    if (!sol)
      return;
    
    ModelSpec *spec = const_cast<ModelSpec *>(&sol->modspec);

    delete sol;
    delete spec;

    R_ClearExternalPtr(ptr);
  }

  SEXP
  multi_country_solution(SEXP n, SEXP solution_method, SEXP order)
  {
    PROTECT(n = coerceVector(n, INTSXP));
    PROTECT(solution_method = coerceVector(solution_method, STRSXP));
    PROTECT(order = coerceVector(order, INTSXP));

    MultiCountryRBCA2 *spec = new MultiCountryRBCA2(*INTEGER(n), 0.36, 0.99,
                                                    0.025, 0.95, 0.01, 0.5,
                                                    0.25, 1, 0.1, 1);

    struct rusage r1, r2;
    getrusage(RUSAGE_SELF, &r1);

    ModelSolution *sol;
    
    const char *sm = CHAR(STRING_ELT(solution_method, 0));
    if (!strcmp(sm, "cga"))
      sol = new MultiCountryCGASolution(*spec);
    else
      {
        MultiCountryHybSolution::hybrid_t m;
        if (!strcmp(sm, "per"))
          m = MultiCountryHybSolution::PlainPerturbation;
        else if (!strcmp(sm, "hyb"))
          m = MultiCountryHybSolution::Hybrid;
        else
          error_return("Unknown solution method");

        try
          {
            sol = new MultiCountryHybSolution(*spec, *INTEGER(order), m);
          }
        catch (ModelSolution::UnsupportedSpec &e)
          {
            error_return("Unsupported specification");
          }
      }

    getrusage(RUSAGE_SELF, &r2);
    SEXP sol_time;
    PROTECT(sol_time = allocVector(REALSXP, 1));
    *REAL(sol_time) = cputime(r1, r2);

    SEXP ptr;
    PROTECT(ptr = R_MakeExternalPtr(sol, R_NilValue, R_NilValue));
    R_RegisterCFinalizerEx(ptr, solution_finalizer, TRUE);

    SEXP r;
    PROTECT(r = allocVector(VECSXP, 2));
    SET_VECTOR_ELT(r, 0, ptr);
    SET_VECTOR_ELT(r, 1, sol_time);

    SEXP names;
    PROTECT(names = allocVector(STRSXP, 2));
    SET_STRING_ELT(names, 0, mkChar("sol"));
    SET_STRING_ELT(names, 1, mkChar("sol.time"));
    namesgets(r, names);

    UNPROTECT(7);

    return r;
  }
}
